/*  SH95UPD v0.0.6 by SP193
 *  BUILD 090401
 *
 *  KernelEx
 *  Copyright (C) 2006-2007, Xeno86
 *
 *  Original WIN98UPDT0001
 *  Copyright (C) 2005/2006, (x)
 *
 *  This file is part of KernelEx source code.
 *
 *  KernelEx is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; version 2 of the License.
 *
 *  KernelEx is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/** @file main.cpp
 * SH95UPD Setup program.
 */

#ifdef __BORLANDC__
#pragma resource "SH95UPD.RES"
#endif

#include <stdio.h>
#include <windows.h>
#include "pefile.h"
#include "common.h"
#include "wininit.h"

#include "vrprintf.c"

#define RECOVERY_FILE "SH95RECO.BAT"
#define UNINST_PREFIX "SH95UPD"

extern int shell32_updt(CPEFile &);

char SH95VER[7];

//---------------------------------------------------------------------------

/** Reads the version of SH95UPD from SH95UPD.dat */
int readSH95ver() {

		char setupdata[MAX_PATH];
		GetSystemDirectory(setupdata, MAX_PATH);
		strcat(setupdata, "\\" UNINST_PREFIX ".dat");
		vrprintf(2, "Checking for %s ... ", UNINST_PREFIX ".dat");
		if (FileExists(setupdata))
		{
			FILE * uninst_file = fopen(setupdata, "rt");
			fscanf(uninst_file, "%s\n", SH95VER);
			fclose(uninst_file);
			vrprintf(2, "Version %s found\n", SH95VER);
			return 1;
	}
	return 0;		
}

//---------------------------------------------------------------------------
/** Holds information about available patcher modules. */
static struct
{
	char orig_name[20];     /**< Module file name. */
	char bak_name[20];      /**< Backup file name. */
	bool optional;          /**< Is the module optional? */
	int (* updt_func)(CPEFile &); /**< Address of module's update function. */
	int (* aux_fun)(void);  /**< Address of auxiliary function. Called after module's update function. */
} FilesToUpdate[] = {
	{ "SHELL32.DLL",          "SHELL32.S95",          false,  shell32_updt,   NULL          },
	{ "",                     "",                     false,  NULL,           NULL          }
};

static bool param_remove = 0;
static bool param_reinstall = 0;
static bool param_batch = 0;
static bool param_log = 1;
static bool param_verify = 0;

//pointer to SH95Version function
typedef void (WINAPI *KV_t)(char *);
static KV_t h_KVersion;     /**< Handle to SH95Version function. */

unsigned long MinorVersion; /**< Minor Windows version. */

//---------------------------------------------------------------------------

/** Overloaded MessageBox function.
 * @param txt Message.
 * @param type MessageBox flags.
 * @return Same as original MessageBox.
 */
int MessageBox(const char *txt, unsigned int type)
{
	return MessageBox(NULL, txt, "SH95UPD", type);
}

/** Displays an error message and quits.
 * @param msg Message.
 * @param error_code Error code. Quit program if not zero.
 */
void PushError(const char *msg, int error_code)
{
	vrprintf(0, "ERROR! %s\n", msg);
	if (!param_batch) MessageBox(msg, MB_OK | MB_ICONERROR);
	if (error_code) exit(error_code);
}

//---------------------------------------------------------------------------

/** Processes command line parameters.
 * @param argc Argument count.
 * @param argv Argument vector.
 */
void process_commandline(int argc, char *argv[])
{
	for (int i = 0 ; i < argc ; i++)
	{
		if ((argv[i][0] == '/') || (argv[i][0] == '-'))
		{
			if		 (!strcmp(argv[i]+1, "remove")) param_remove = 1;
			else if (!strcmp(argv[i]+1, "debug-reinstall")) param_reinstall = 1; //this is for debugging purposes ONLY!!
			else if (!strcmp(argv[i]+1, "v0")) vrSetVerboseLevel(0);
			else if (!strcmp(argv[i]+1, "v1")) vrSetVerboseLevel(1);
			else if (!strcmp(argv[i]+1, "v2")) vrSetVerboseLevel(2);
			else if (!strcmp(argv[i]+1, "v3")) vrSetVerboseLevel(3);
			else if (!strcmp(argv[i]+1, "batch")) param_batch = 1;
			else if (!strcmp(argv[i]+1, "nolog")) param_log = 0;
			else if (!strcmp(argv[i]+1, "log")) param_log = 1;
			else if (!strcmp(argv[i]+1, "verify")) param_verify = 1;
			else
			{
				char err[100];
				sprintf(err, "Invalid parameter: '%s'", argv[i]);
				PushError(err, 1);
			}
		}
	}
}

//---------------------------------------------------------------------------

/** Patcher modules can call this function to obtain the flags.
 * @return Or'ed value: CMD_REMOVE - uninstall mode; CMD_REINSTALL - reinstall mode; CMD_BATCH - batch mode.
 */
int get_cmdline_params(void)
{
	return (param_remove ? CMD_REMOVE : 0) | (param_reinstall ? CMD_REINSTALL : 0) | (param_batch ? CMD_BATCH : 0);
}

//---------------------------------------------------------------------------
/** Generates temporary file name (full path).
 * @param nn Output buffer (length must be at least MAX_PATH).
 */
void GenTempFileName(char * nn)
{
	char tempdir[MAX_PATH];
	GetSystemDirectory(tempdir, MAX_PATH - 12);
	GetTempFileName(tempdir, "SH95", 0, nn);
	vrprintf(2, "Saving temporary file to: %s\n", nn);
}

//---------------------------------------------------------------------------
/** Verifies SH95UPD installation. */
void verify_installation(void)
{
	vrprintf(0, "Verify mode\n\n");

	int fail_cnt = 1;
	int verreadstatus=0;

	char tmp[MAX_PATH];
	GetSystemDirectory(tmp, MAX_PATH);
	strcat(tmp, "\\" UNINST_PREFIX ".dat");
	vrprintf(1, "Checking for %s ... ", UNINST_PREFIX ".dat");
	if (FileExists(tmp))
	{
		vrprintf(1, "ok\n");
		vrprintf(1, "Trying to open %s ... ", UNINST_PREFIX ".dat");
		FILE * uninst_file = fopen(tmp, "rt");
		if (uninst_file)
		{
			vrprintf(1, "ok\n");

			//verify recovery file
			vrprintf(1, "Checking for recovery file ... ");
			char BatFile[MAX_PATH];
			GetWindowsDirectory(BatFile, MAX_PATH);
			strcat(BatFile, "\\COMMAND\\" RECOVERY_FILE);
			if (FileExists(BatFile)) vrprintf(1, "ok\n");
			else vrprintf(1, "not found\n");

			//verify version
			vrprintf(1, "Checking version ... ");
			verreadstatus=readSH95ver();
			if (!verreadstatus) vrprintf(1, "error\n");
			else
			{
				vrprintf(1, "%s ... ", SH95VER);
				if (!strcmp(SH95VER, VERSION)) {
                                 vrprintf(1, "ok\n");
                                 fail_cnt--;
                                 }
				else vrprintf(1, "mismatch\n");
			}

			//check for executable
			vrprintf(1, "Checking for SH95UPD executable ... ");
			GetSystemDirectory(tmp, MAX_PATH);
			strcat(tmp, "\\" UNINST_PREFIX ".exe");
			if (FileExists(tmp)) vrprintf(1, "ok\n");
			else {
                 vrprintf(1, "not found\n");
                 fail_cnt++;
                 }

			//verify checksums
			char SysFile[MAX_PATH];
			char BakFile[MAX_PATH];
			unsigned long Checksum1, Checksum2;
			int auxidx;
			while (fscanf(uninst_file, "%08lx %d %s %s\n", &Checksum1, &auxidx, SysFile, BakFile) == 4)
			{
				if (strcmp(SysFile, "NUL"))
				{
					strcpy(tmp, SysFile);
					char *t = strrchr(SysFile, '.');
					if (t) *t = '\0';
					t = strrchr(SysFile, '\\');
					if (t) t++; else t = SysFile;
					vrprintf(1, "Checking %s ... ", t);
				}
				else
				{
					strcpy(tmp, BakFile);
					char *t = strrchr(BakFile, '.');
					if (t) *t = '\0';
					t = strrchr(BakFile, '\\');
					if (t) t++; else t = BakFile;
					vrprintf(1, "Checking %s ... ", t);

				}

				if (FileExists(tmp))
				{
					if (Checksum1)
					{
						CPEFile file;
						if (file.LoadFromFile(tmp)) Checksum2 = file.CalculateChecksum();
						else Checksum2 = 0;

						if (Checksum1 == Checksum2)
						{
							vrprintf(1, "ok\n");
						}
						else {
                             vrprintf(1, "mismatch\n");
                             fail_cnt++;
                             }
					}
					else vrprintf(1, "ok\n");
				}
				else { vrprintf(1, "not found\n"); fail_cnt++; }

			}
		}
		else vrprintf(1, "failed\n");
	}
	else vrprintf(1, "failed\n");

	if (fail_cnt) vrprintf(0, "\nErrors were detected!!\nWARNING: SH95UPD may not be installed properly!\n");
	else vrprintf(0, "\nSH95UPD seems to be installed properly.\n");
	if (!param_batch)
	{
		printf("\nPress ENTER to close this window.\n");
		getc(stdin);
	}
	if (!fail_cnt) exit(0);
	else exit(8);
}

//---------------------------------------------------------------------------

/** Uninstalls SH95UPD. */
void uninstall(void)
{
	char SysFile[MAX_PATH];
	char BakFile[MAX_PATH];
	unsigned long Checksum1, Checksum2;
	int auxidx; //index of auxiliary function (VERSION DEPENDANT!!!)

	int fail_cnt = 0;

	HMODULE h_Kernel32 = GetModuleHandle("kernel32.dll");
   h_KVersion = (KV_t)GetProcAddress(h_Kernel32, "KEXVersion"); // for KernelEX compatibility
    if (!h_KVersion) h_KVersion = (KV_t)GetProcAddress(h_Kernel32, "KUPVersion"); // for KUP compatibility
       if (h_KVersion) PushError("Detected existing KernelEX/KUP installation.\n Please uninstall KernelEX/KUP before uninstalling SH95UPD.", 9);

	FILE * uninst_file;
	GetSystemDirectory(SysFile, MAX_PATH);
	strcat(SysFile, "\\" UNINST_PREFIX ".dat");
	Wininit.add("NUL", SysFile);
	vrprintf(1, "Removing %s ... ok\n", UNINST_PREFIX ".dat");
	uninst_file = fopen(SysFile, "rt");
	fscanf(uninst_file, "%s\n", SysFile); //skip version number

	while (fscanf(uninst_file, "%08lx %d %s %s\n", &Checksum1, &auxidx, SysFile, BakFile) == 4)
	{
		if (strcmp(SysFile, "NUL") != 0)
		{
			char *t = strrchr(SysFile, '\\');
			if (t) t++; else t = SysFile;
			vrprintf(1, "Restoring %s ... ", t);
		}
		else
		{
			char *t = strrchr(BakFile, '\\');
			if (t) t++; else t = BakFile;
			vrprintf(1, "Removing %s ... ", t);
		}


		if (strcmp(SysFile, "NUL") != 0) //system files
		{
			if (FileExists(SysFile))
			{
				CPEFile file;

				if (file.LoadFromFile(SysFile)) Checksum2 = file.CalculateChecksum();
				else Checksum2 = 0;

				if (!Checksum1 || Checksum1 == Checksum2)
				{
					if (FileExists(BakFile))
					{
						Wininit.add(SysFile, BakFile);
						vrprintf(1, "ok\n");
					}
					else
					{
						vrprintf(1, "file missing\n");
						if (!param_batch) MessageBox("Couldn't find backup file!! Use SFC to fix", MB_OK | MB_ICONERROR);
						fail_cnt++;
					}
				}
				else
				{
					Wininit.add("NUL", BakFile);
					vrprintf(1, "checksum mismatch - skipping\n");
				}
			}
			else
			{
				if (FileExists(BakFile))
				{
					Wininit.add(SysFile, BakFile);
					vrprintf(1, "ok\n");
				}
				else
				{
					vrprintf(1, "invalid entry\n");
				}
			}
		}
		else //custom files
		{
			if (FileExists(BakFile))
			{
				CPEFile file;

				if (file.LoadFromFile(BakFile)) Checksum2 = file.CalculateChecksum();
				else Checksum2 = 0;

				if (!Checksum1 || Checksum1 == Checksum2)
				{
					Wininit.add("NUL", BakFile);
					vrprintf(1, "ok\n");
				}
				else
				{
					vrprintf(1, "checksum mismatch - skipping\n");
				}
			}
			else
			{
				vrprintf(1, "invalid entry\n");
			}
		}

		if (auxidx >= 0 && FilesToUpdate[auxidx].aux_fun) FilesToUpdate[auxidx].aux_fun();
	}

	fclose(uninst_file);


	//remove executable
	vrprintf(1, "Removing SH95UPD executable ... ");
	GetSystemDirectory(SysFile, MAX_PATH);
	strcat(SysFile, "\\" UNINST_PREFIX ".exe");
	if (FileExists(SysFile))
	{
		Wininit.add("NUL", SysFile);
		vrprintf(1, "ok\n");
	}
	else
	{
		vrprintf(1, "file missing\n");
		fail_cnt++;
	}

	//remove registry entry
	vrprintf(2, "Removing uninstaller registry entry\n");
	RegDeleteKey(HKEY_LOCAL_MACHINE, "Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\" UNINST_PREFIX);


	if (!fail_cnt)
	{
#ifdef RECOVERY_FILE
		//remove RECOVERY_FILE
		char BatFile[MAX_PATH];
		GetWindowsDirectory(BatFile, MAX_PATH);
		strcat(BatFile, "\\COMMAND\\" RECOVERY_FILE);
		vrprintf(1, "Removing recovery file: %s\n", BatFile);
		Wininit.add("NUL", BatFile);
#endif
		Wininit.save();
		if (param_batch) vrprintf(0, "SH95UPD removed - please reboot to finish uninstallation\n");
		else MessageBox("To complete the uninstallation, you must restart your computer.",MB_OK|MB_ICONINFORMATION);
	}
	else
	{
		Wininit.save();
		PushError("One or more backup files were missing!!\n"
				"Restore those files manually using SFC\nPlease reboot to apply changes", 5);
	}
	exit (fail_cnt != 0);
}

//---------------------------------------------------------------------------

/** Installs SH95UPD. */
void install(void)
{
	//////////////////// INSTALLATION ////////////////////


	//file path name variables
	char SysDll[MAX_PATH];
	char BakFile[MAX_PATH];

	int skip = 0;
	int errors = 0;

	//Create uninstaller data file
	GenTempFileName(BakFile);
	FILE * uninst_file = fopen(BakFile, "wt");
	GetSystemDirectory(SysDll, MAX_PATH);
	strcat(SysDll, "\\" UNINST_PREFIX ".dat");
	Wininit.add(SysDll, BakFile);

	//insert version number
	fprintf(uninst_file, "%s\n", VERSION);
	
	vrprintf(1, "\n");

	//file patching loop
	for (int i = 0 ; FilesToUpdate[i].orig_name[0] != '\0' ; i++)
	{
		vrprintf(0, "\n");

		GetSystemDirectory(SysDll, MAX_PATH);
		strcat(SysDll, "\\");
		strncpy(BakFile, SysDll, MAX_PATH);

		strcat(SysDll, FilesToUpdate[i].orig_name);
		strcat(BakFile, FilesToUpdate[i].bak_name);

		//create file object
		CPEFile File2Patch;

		if (param_reinstall)
		{
			//check if backup exists
			if (FileExists(BakFile))
			{
				//try to load file
				if (!File2Patch.LoadFromFile(BakFile, PATCH_MAX_SIZE))
				{
					char err[100];
					sprintf(err, "Couldn't open file: %s !! Skipping", BakFile);
					PushError(err, 0);
					skip++;
					continue;
				}
				vrprintf(0, "Reinstalling...\n");
			}
			else
			{
				char err[100];
				sprintf(err, "Couldn't find file: %s !! Skipping", BakFile);
				PushError(err, 0);
				skip++;
				continue;
			}
		}
		else
		{
			// try to load file
			if (!File2Patch.LoadFromFile(SysDll, PATCH_MAX_SIZE))
			{
				vrprintf(1, "ERROR! Couldn't open file: %s !! Aborting\n", SysDll);
				errors++;
				break;
			}

			//check if patch was already applied
			if (File2Patch.GetSection(UPDT_SEC_NAME))
			{
				char err[100];
				sprintf(err, "File %s contains a conflicting modification!", FilesToUpdate[i].orig_name);
				PushError(err, 0);
				errors++;
				break;
			}

			vrprintf(0, "Patching...\n");
		}

		//show file info
		if (!param_reinstall)
		{
			vrprintf(1, "Backup file: %s\n", BakFile);
			vrprintf(1, "File to patch: %s\n", SysDll);
		}
		else vrprintf(1, "File to patch: %s\n", BakFile);

		//execute patcher
		if (FilesToUpdate[i].updt_func(File2Patch) != PATCH_OK)
		{
			if (FilesToUpdate[i].optional)
			{
				vrprintf(1, "Optional patch skipped\n");
				continue;
			}
			else
			{
				errors++;
				vrprintf(1, "Patch failed\n");
				break;
			}
		}
		vrprintf(1, "Patch succeeded\n");

		//fix checksum
		File2Patch.FixChecksum();

		//create backup
		if (!param_reinstall)
		{
			char tmp_name[MAX_PATH];
			GenTempFileName(tmp_name);
			CopyFile(SysDll, tmp_name, FALSE);
			Wininit.add(BakFile, tmp_name);
		}

		//save temporary file
		char tmp_name[MAX_PATH];
		GenTempFileName(tmp_name);
		File2Patch.SaveToFile(tmp_name);
		Wininit.add(SysDll, tmp_name);

		//perform additional tasks
		if (FilesToUpdate[i].aux_fun)
		{
			vrprintf(1, "Performing auxiliary task...\n");
			if (FilesToUpdate[i].aux_fun() == PATCH_FAIL)
			{
				errors++;
				vrprintf(1, "Aux task failed\n");
				break;
			}
			vrprintf(1, "Aux task succeeded\n");
		}

		//calculate checksum and add entry to uninstall log
		unsigned long Checksum = File2Patch.CalculateChecksum();
		fprintf(uninst_file, "%08lx %d %s %s\n", Checksum, FilesToUpdate[i].aux_fun ? i : -1, SysDll, BakFile);
	}

	fclose(uninst_file);

#ifdef RECOVERY_FILE
	//create recovery file
	if (!errors && !param_reinstall)
	{
		char BatFile[MAX_PATH];
		GetWindowsDirectory(BatFile, MAX_PATH);
		strcat(BatFile, "\\COMMAND\\" RECOVERY_FILE);

		vrprintf(1, "\nCreating recovery file: %s\n", BatFile);

		FILE * f = fopen(BatFile, "wt");
		fprintf(f,
			"@echo off\n"
			"echo SH95UPD Recovery Script\n"
			"set SH95PATH=\n"
			"if not \"%%1\"==\"\" set SH95PATH=%%1\n"
			"if not \"%%SH95PATH%%\"==\"\" goto have_path\n"
			"if not \"%%WINDIR%%\"==\"\" set SH95PATH=%%WINDIR%%\n"
			"if not \"%%SH95PATH%%\"==\"\" goto have_path\n"
			"if exist C:\\WINDOWS\\nul set SH95PATH=C:\\WINDOWS\n"
			"if not \"%%SH95PATH%%\"==\"\" goto have_path\n"
			"\n"
			"echo Couldn't find path to Windows\n"
			"echo Type: %%0 drive:\\yourpathtowindows\n"
			"echo e.g.: %%0 C:\\WINDOWS\n"
			"goto end\n"
			"\n"
			":have_path\n"
			"set SH95PATH=%%SH95PATH%%\\SYSTEM\n"
			"echo Searching for backup files in %%SH95PATH%%\n"
			"if not exist %%SH95PATH%%\\*.S95 goto missing\n"
		);
		for (int i = 0 ; FilesToUpdate[i].orig_name[0] != '\0' ; i++)
		{
			fprintf(f,
				"if exist %%SH95PATH%%\\%s del %%SH95PATH%%\\%s\n"
				"if exist %%SH95PATH%%\\%s copy /B %%SH95PATH%%\\%s %%SH95PATH%%\\%s\n",
				FilesToUpdate[i].bak_name,
				FilesToUpdate[i].orig_name,
				FilesToUpdate[i].bak_name,
				FilesToUpdate[i].bak_name,
				FilesToUpdate[i].orig_name
			);
		}
		fprintf(f,
			"goto done\n"
			"\n"
			":missing\n"
			"echo Error! Backup file(s) missing!\n"
			"goto end\n"
			"\n"
			":done\n"
			"echo All done!\n"
			"\n"
			":end\n"
		);
		fclose(f);
	}
#endif


	//////////////////// ERROR PROCESSING ////////////////////


	//do clean up and show pending error messages
	if (errors)
	{
		param_remove = 1;

		for (int i = 0 ; FilesToUpdate[i].orig_name[0] != '\0' ; i++)
		{
			if (FilesToUpdate[i].aux_fun) FilesToUpdate[i].aux_fun();
		}

		PushError("Some of the patches failed!!\nInstallation aborted!!", 8);
	}
	//message for reinstall
	if (skip) PushError("Some of the files were missing!! Please check your system.", 5);


	//////////////////// SUCCESSFUL INSTALLATION ////////////////////


	//save wininit, create uninstaller
	char tmp[MAX_PATH];
	GetModuleFileName(NULL, BakFile, MAX_PATH);
	GenTempFileName(tmp);
	CopyFile(BakFile, tmp, FALSE);
	GetSystemDirectory(BakFile, MAX_PATH);
	strcat(BakFile, "\\" UNINST_PREFIX ".exe");
	Wininit.add(BakFile, tmp);

	Wininit.save();

	//add verifier to system statup for next boot
	HKEY hkey;
	bool fail = true;
	if (!param_batch)
	{
		if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, "Software\\Microsoft\\Windows\\CurrentVersion\\RunServicesOnce", 0, KEY_SET_VALUE, &hkey) == ERROR_SUCCESS)
		{
			strcpy(BakFile, UNINST_PREFIX ".exe -verify -nolog");
			if (RegSetValueEx(hkey, "SH95UPD", 0, REG_SZ, (BYTE *)BakFile, strlen(BakFile) + 1) == ERROR_SUCCESS)
			{
				vrprintf(2, "Added verifier registry entry\n");
				fail = false;
			}
			RegCloseKey(hkey);
		}
		if (fail) vrprintf(2, "Failed to add verifier registry entry\n");
	}

	DWORD disp;
	fail = true;
	if (RegCreateKeyEx(HKEY_LOCAL_MACHINE, "Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\" UNINST_PREFIX, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL, &hkey, &disp) == ERROR_SUCCESS)
	{
		char tmp[64];
		strcpy(tmp, "SH95UPD v");
		strcat(tmp, VERSION);
		if (RegSetValueEx(hkey, "DisplayName", 0, REG_SZ, (BYTE *)tmp, strlen(tmp) + 1) == ERROR_SUCCESS)
		if (RegSetValueEx(hkey, "UninstallString", 0, REG_SZ, (BYTE *)UNINST_PREFIX ".exe -remove", sizeof(UNINST_PREFIX ".exe -remove")) == ERROR_SUCCESS)
		{
			vrprintf(2, "Added uninstaller registry entry\n");
			fail = false;
		}
		RegCloseKey(hkey);
	}
	if (fail) vrprintf(2, "Failed to add uninstaller registry entry\n");

	/* reboot */
	if (param_batch) vrprintf(0, "Patching completed successfully. Please reboot to apply upgrade.\n");
	else MessageBox("For the installation to be completed, you must restart your computer.",MB_OK|MB_ICONINFORMATION);
}

//---------------------------------------------------------------------------
/** Processes command line, performs checks, sets operating mode.
 * @param argc Argument count.
 * @param argv Argument vector.
 * @return Exit code.
 */
int main(int argc, char *argv[])
{
	int verreadstatus=0;
	vrSetVerboseLevel(1);
	process_commandline(argc, argv);

	if (param_log)
	{
		char LogFile[MAX_PATH];
		strcpy(LogFile, argv[0]);
		LogFile[strlen(LogFile) - 4] = '\0';
		strcat(LogFile, ".log");
		vrEnableLog(LogFile);
		atexit(vrDisableLog);
		atexit(WininitClear);
	}

    vrprintf(0, "SH95UPD version %s by SP193\n", VERSION);
	vrprintf(0, "Based on KernelEx version 0.3.2a\n");
	vrprintf(0, "Formerly Kernel Update Project.\nCopyright (C) 2006-2007, Xeno86.\n\n");

	if (param_verify) verify_installation();


	//////////////////// CHECKS ////////////////////


	//check if is windows 9x
	OSVERSIONINFO WinVersion;
	WinVersion.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	if (!GetVersionEx(&WinVersion))
	{
		PushError("GetVersionEx() failed!!", 2);
	}

	vrprintf(2, "Detected Windows version: %d.%d.%d\n", WinVersion.dwMajorVersion, WinVersion.dwMinorVersion, LOWORD(WinVersion.dwBuildNumber));

	if ( (WinVersion.dwPlatformId != VER_PLATFORM_WIN32_WINDOWS) || (WinVersion.dwMinorVersion != 0 && WinVersion.dwMinorVersion != 10 && WinVersion.dwMinorVersion != 90) ) // Windows 95 compatible
	{
		PushError("Incompatible Windows version!! Windows 95/98 or ME required", 2);
	}
	else MinorVersion = WinVersion.dwMinorVersion;

	//prevent running if WININIT.INI exists
	if (Wininit.exists())
	{
		PushError("Detected 'WININIT.INI' file! Can't proceed! Please reboot first!", 4);
	}

	//show warning for WinME users
	if (!param_batch && MinorVersion == 90)
	{
		if (MessageBox("You are using Windows ME\nMake sure to disable System File Protection feature before continuing (see NOTES)\nProceed?", MB_OKCANCEL | MB_ICONWARNING) == IDCANCEL)
			PushError("Aborted by user!!", 3);
	}

	//detect previous KernelEX installation
	HMODULE h_Kernel32 = GetModuleHandle("kernel32.dll");
   h_KVersion = (KV_t)GetProcAddress(h_Kernel32, "KEXVersion"); // for KernelEX compatibility
    if (!h_KVersion) h_KVersion = (KV_t)GetProcAddress(h_Kernel32, "KUPVersion"); // for KUP compatibility
      if (!h_KVersion) h_KVersion = (KV_t)GetProcAddress(h_Kernel32, "SH95Version"); /* For backward compatibility */
	if (!h_KVersion) verreadstatus=readSH95ver();
	 if ((verreadstatus || h_KVersion) && !param_reinstall)
	{
		char buf[100];
		if (h_KVersion) h_KVersion(buf);
		else strcpy(buf,SH95VER);

		if (strcmp(buf, VERSION))
		{
			char err[120];
			sprintf(err, "Incompatible version is already installed: %s !!\nUninstall it before proceeding", buf);
			PushError(err, 9);
		}
	}

	//check uninstaller data file
	if (!param_reinstall)
	{
		char setupdata[MAX_PATH];
		GetSystemDirectory(setupdata, MAX_PATH);
		strcat(setupdata, "\\" UNINST_PREFIX ".dat");
		vrprintf(2, "Checking for %s ... ", UNINST_PREFIX ".dat");
		if (FileExists(setupdata))
		{
			char buf[100];
			FILE * uninst_file = fopen(setupdata, "rt");
			fscanf(uninst_file, "%s\n", buf);
			fclose(uninst_file);
			vrprintf(2, "found\n");
			//if version differs
			if (strcmp(buf, VERSION))
			{
				char err[120];
				sprintf(err, "Another version is already installed: %s !!\nUninstall it before proceeding", buf);
				PushError(err, 9);
			}
			//if version equals
			else if ((param_batch && param_remove) || (param_remove && MessageBox("Proceed with UNINSTALLATION?", MB_OKCANCEL | MB_ICONQUESTION) == IDOK)
				|| (!param_remove && MessageBox("SH95UPD is already installed\nWould you like to UNINSTALL?", MB_OKCANCEL | MB_ICONINFORMATION) == IDOK))
			{
				vrprintf(0, "\nSwitching to Uninstallation mode\n\n");
				param_remove = 1;
				uninstall();
			}
			else PushError("Uninstallation aborted by user!!", 3);
		}
		else vrprintf(2, "not found\n");
	}

	//show questions and print info about selected mode
	if (param_remove)
	{
		PushError("Not installed!!", 1);
	}
	else if (param_reinstall)
	{
		if (!param_batch && MessageBox("Proceed with reinstallation?", MB_OKCANCEL | MB_ICONQUESTION) == IDCANCEL)
			PushError("Reinstallation aborted by user!!", 3);
		vrprintf(0, "Reinstallation mode\n");
	}
	else
	{
		vrprintf(0, "Installation mode\n");
	}


	//check if the user knows what is doing
	if (!param_batch && !param_reinstall &&
		MessageBox("This software is provided \"as-is,\" without any express or implied warranty.\n"
					"In no event shall the author be held liable for any damages arising from the\n"
					"use of this software.\n\n"
					"By clicking 'OK' you agree with KernelEx license terms.\n"
					"Do you want to install SH95UPD?", MB_OKCANCEL | MB_ICONQUESTION)
		 == IDCANCEL)
	{
		PushError("Installation aborted by user!!", 3);
	}

	install();
	return 0;
}
