/*
 *  KernelEx
 *  Copyright (C) 2006, Xeno86
 *
 *  This file is part of KernelEx source code.
 *
 *  KernelEx is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; version 2 of the License.
 *
 *  KernelEx is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/** @file patch.h
 * Class patch_item - declaration.
 */

#ifndef __PATCH_H
#define __PATCH_H

#include "common.h"

#define API_ENTRY_NAME_LEN      44

// class declarations

/** @struct api_entry
 * Holds an offset and function name.
 */
struct api_entry
{
	DWORD addr;
	char name[API_ENTRY_NAME_LEN];
};

/** @class patch_item
 * A class providing automated manipulation of binary patches.
 */
class patch_item
{
protected:
	char * code;            /**< The code used in the patch. */
	int code_size;          /**< Size of the code. */
	CPEFile * Destination;  /**< Target PE file. */
	DWORD ImgBase;          /**< Image base of the PE file. */
private:
	DWORD * abs_ofs;        /**< Absolute pointers. Table containing offsets to data which needs fixing addresses and relocation. */
	api_entry * abs_api;    /**< Absolute indirect calls and jumps. Table containing offsets and names of external functions (imported by PE file; needing relocation). */
	api_entry * rel_api;    /**< Relative calls and jumps. Table containing offsets and names of internal functions (exported by PE file). */
	api_entry * hook_api;   /**< Hooks. Table containing offsets and names of functions we want to export or replace (if already present in PE file). */
	int (* custom)();       /**< Optional function called by Process() before copying the code to PE file. */
public:
	patch_item(char * binary_code, int size, DWORD abs_ofs[], api_entry abs_api[], api_entry rel_api[], api_entry hook_api[], int (* custom)());
	virtual ~patch_item();
	int virtual GetPatchSize();
	int GetRelocSize(DWORD Rva, DWORD & prevBlRva, int & lastblocksize);
	int SetWorkFile(CPEFile * dest);
	int AddExports();
	int virtual Process(DWORD Rva, char * ptr);
};

/** @class patch_item_alias
 * A dummy-patch class creating an alias over existing export.
 */
class patch_item_alias : public patch_item
{
private:
	api_entry hook_api[2];  /**< Contains target function name. */
	char existing_export[API_ENTRY_NAME_LEN]; /**< Existing function name. */
public:
	patch_item_alias(const char * new_name, const char * existing_name);
	int Process(DWORD Rva, char * ptr);
};

// function declarations
int apply_patches(CPEFile & File2Patch, patch_item * patches[]);

#endif
