/*  SH95UPD v0.0.2A
 *
 *
 *  KernelEx
 *  Copyright (C) 2006, Xeno86
 *
 *  Original WIN98UPDT0001
 *  Copyright (C) 2005/2006, (x)
 *
 *  This file is part of KernelEx source code.
 *
 *  KernelEx is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; version 2 of the License.
 *
 *  KernelEx is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/** @file pefile.h
 * Class CPEFile - declaration.
 */

#ifndef __PEFILE_H
#define __PEFILE_H

#include <windows.h>
#include <winnt.h>
#include <string>
#include <vector>
#include "File.h"

//flags for RebuildExportTable()
#define E_ADDORINC      0
#define E_MUSTCREATE    1

/** @class CPEFile
 * Class providing basic operations on Portable Executable (PE) files.
 */
class CPEFile : public CFile
{

protected:
	IMAGE_DOS_HEADER* m_MZHeader;                    /**< Pointer to MS-DOS header (start of the file). */
	IMAGE_NT_HEADERS* m_PEHeader;                    /**< Pointer to PE header. */
	IMAGE_SECTION_HEADER* m_SectionHeaders;          /**< Pointer to first section header. */

	std::vector<std::string> m_NewExports;           /**< Holds names of exports to add to PE file. */
	bool BuildExportTable(void* Raw, int Rva);

	std::vector<DWORD> m_NewRelocs;                  /**< Holds RVAs of data that needs relocation. */
	DWORD m_RelocTableCapacity;                      /**< Space reserved by AllocRelocTable for new entries. @see AllocRelocTable */
public:
	CPEFile();
	virtual ~CPEFile();

	bool LoadFromFile(const char* FilePath, int SizeToAdd = 0);
	void* AddSection(const char* SectionName, int SectionSize, int* NewSectionRVA, DWORD Characteristics = 0);
	void* IncSection(IMAGE_SECTION_HEADER* SectionHeader, int AddedSectionSize, int* NewSectionRVA);

	void SetEntryPoint(int NewEntryPoint);

	void* GetEntryPoint();
	void* GetImageBase();
	IMAGE_NT_HEADERS* GetPEHeader();
	IMAGE_DOS_HEADER* GetMZHeader();

	virtual void* RvaToOffset(DWORD Rva);

	IMAGE_SECTION_HEADER* GetSection(int Section);
	IMAGE_SECTION_HEADER* GetSection(const char* SectionName);

	void* GetApiImport(const char* Library, const char* Function);
	void* GetApiImport(const char* Library, WORD Ordinal);

	void* GetApiExport(const char* Function, int* ApiPtr = 0);
	void* GetApiExport(int Ordinal);

	bool ChangeImageBase(int NewBase);

	DWORD CalculateChecksum();
	void FixChecksum();

	void AddExport(const char* ExportName);
	int GetExportTableSize();
	bool RebuildExportTable(const char* mtd2secname, int Flags = E_ADDORINC);
	bool RebuildExportTable_MTD2(const char* SectionName, int Flags = E_MUSTCREATE);

	IMAGE_SECTION_HEADER* RvaToSection(DWORD Rva);
	bool RebuildExportTable_MTD1();
	void AddReloc(DWORD Rva);
	bool AllocRelocTable(int BytesToAdd);
	bool UpdateRelocTable();
	void RemoveDebug();
	DWORD GetApiExportNameRVA(const char* Function);
};

#endif
