/*
 *  KernelEx
 *  Copyright (C) 2006-2007, Xeno86
 *
 *  This file is part of KernelEx source code.
 *
 *  KernelEx is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; version 2 of the License.
 *
 *  KernelEx is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/** @file vrprintf.c
 * Verbose printf - implementation.
 */

#include <stdio.h>
#include <stdarg.h>

/** Current verbosity level. */
static int VerboseLevel = 0;

/** Handle for logging to file. */
static FILE *handle = NULL;

/** Sets current verbosity level to a new value.
 * @param lev Desired verbosity level.
 */
void vrSetVerboseLevel(int lev)
{
	VerboseLevel = lev;
}

/** Prints debug message.
 * @param MinVerboseLevel Minimum verbosity level required to show the message on screen.
 *        The higher, the more info is displayed.
 *        NOTE: File has VerboseLevel + 1.
 * @param fmt Format specifier.
 * @return Number of bytes output.
 */
int vrprintf(int MinVerboseLevel, const char *fmt, ...)
{
	va_list argp;
	int cnt = 0;
	if (VerboseLevel + 1 < MinVerboseLevel) return 0;
	va_start(argp, fmt);
	if (VerboseLevel >= MinVerboseLevel) cnt = vprintf(fmt, argp);
	if (handle)
	{
		vfprintf(handle, fmt, argp);
		fflush(handle);
	}
	va_end(argp);
	return cnt;
}

/** Enables logging to file.
 * @param fn File name.
 */
void vrEnableLog(char *fn)
{
	handle = fopen(fn, "w");
}

/** Disables logging to file. */
void __cdecl vrDisableLog(void)
{
	if (handle) fclose(handle);
	handle = NULL;
}
