/*
 *  GenPatch
 *  Copyright (C) 2006-2007, Xeno86
 *
 *  This file is part of KernelEx source code.
 *
 *  KernelEx is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; version 2 of the License.
 *
 *  KernelEx is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>

#define DUMMY_KEYWORD "DUMMY"

/**************************************************/
/* DEFINICJE DLA KOLEJKI                          */
typedef struct _kolejka
{
    struct _kolejka * next;
    char d1[128];
    char d2[128];
} *kolejka;

void push(kolejka * front, kolejka * back, char * d1, char * d2)
{
    kolejka tmp = (kolejka)malloc(sizeof(struct _kolejka));
    tmp->next = NULL;
    strncpy(tmp->d1, d1, 127);
    strncpy(tmp->d2, d2, 127);

    if (!*back)
    {
        *front = *back = tmp;
    }
    else
    {
        (*back)->next = tmp;
        *back = (*back)->next;
    }
}

void pop(kolejka * front, kolejka * back)
{
    kolejka tmp = *front;
    if (!*front) return;
    *front = (*front)->next;
    free(tmp);
    if (!*front) *back = NULL;
}
/* DEFINICJE DLA KOLEJKI                          */
/**************************************************/

kolejka exports_front;
kolejka exports_back;

kolejka imports_front;
kolejka imports_back;

kolejka calls_front;
kolejka calls_back;


#define VERSION "1.01"
#define TEMP_DIR "temp"

int param_debug = 0;

char confFile[MAX_PATH];
char SrcFile[MAX_PATH];
char SrcFilePrefix[MAX_PATH];
char Ext[4];

//---------------------------------------------------------------------------

int run(char * cmd)
{
    STARTUPINFO si;
    PROCESS_INFORMATION pi;
    DWORD ec;

    memset(&si, 0, sizeof(si));
    si.cb = sizeof(si);

    if (param_debug) printf("Executing command: %s\n", cmd);

    if (CreateProcess(NULL, cmd, NULL, NULL, TRUE, 0 /* CREATE_NEW_CONSOLE */, NULL, NULL/*workdir*/, &si, &pi))
    {
        WaitForSingleObject(pi.hProcess, INFINITE);
        GetExitCodeProcess(pi.hProcess, &ec);
        CloseHandle(pi.hProcess);
    }
    else ec = -1;

    if (param_debug) printf("Exit code: %ld\n", ec);

    return ec;
}

//---------------------------------------------------------------------------

void readini(const char * KeyName, const char * Default, char * Buffer)
{
    GetPrivateProfileString("GenPatch", KeyName, Default, Buffer, 256, confFile);
}

//---------------------------------------------------------------------------
// check for tools
int step0(void)
{
    printf("Step 0 - checking for tools\n");

    // check for tempdir
    if (GetFileAttributes(TEMP_DIR) == 0xFFFFFFFF)
    {
        if (!CreateDirectory(TEMP_DIR, NULL))
        {
            printf("Couldn't create '" TEMP_DIR "' directory!\n");
            return 1;
        }
    }

    char cmd[256];
    FILE * verfile;

    // check for nasm
    readini("NASM_EXEC", "nasm", cmd);
    strcat(cmd, " -v >" TEMP_DIR "\\ver.txt");
    system(cmd);

    verfile = fopen(TEMP_DIR "\\ver.txt", "rt");
    if (!verfile)
    {
        printf("Error: NASM not found!\n");
    }
    fgets(cmd, sizeof(cmd), verfile);
    fclose(verfile);
    if (strcmp(cmd, "NASM version 0.98.39 compiled on Jan 16 2005\n"))
    {
        printf("Error: NASM version does not match!\n");
        return 1;
    }
    remove(TEMP_DIR "\\ver.txt");

    // check for att2intl
    readini("I2G_CMD", "att2intl", cmd);
    strcat(cmd, " -v >" TEMP_DIR "\\ver.txt");
    system(cmd);

    verfile = fopen(TEMP_DIR "\\ver.txt", "rt");
    if (!verfile)
    {
        printf("Error: Att2Intl not found!\n");
        return 1;
    }
    fgets(cmd, sizeof(cmd), verfile);
    fgets(cmd, sizeof(cmd), verfile);
    fclose(verfile);
    if (strcmp(cmd, "Att2Intl v0.2.1a\n"))
    {
        printf("Error: Att2Intl version does not match!\n");
        return 1;
    }
    remove(TEMP_DIR "\\ver.txt");

    return 0;
}
//---------------------------------------------------------------------------
// convert C to S (gas assembly) and convert to intel style
int step1(void)
{
    if (!stricmp(Ext, "asm"))
    {
        char tmp[MAX_PATH];
        strcpy(tmp, TEMP_DIR "\\");
        strcat(tmp, SrcFilePrefix);
        strcat(tmp, ".asm");
        CopyFile(SrcFile, tmp, 0);
        printf("detected asm file - skipping step 1\n");
        return 0;
    }

    printf("Step 1 - converting C to Asm\n");

    char gcc_cmd[256];
    readini("GCC_AS_CMD", "gcc -Os -s -Wall -pedantic -fomit-frame-pointer -march=i386 -mtune=i686 -S", gcc_cmd);
    strcat(gcc_cmd, " -o " TEMP_DIR "\\");
    strcat(gcc_cmd, SrcFilePrefix);
    strcat(gcc_cmd, ".s \"");
    strcat(gcc_cmd, SrcFile);
    strcat(gcc_cmd, "\"");

    if (run(gcc_cmd))
    {
       printf("Error executing GCC command!\n");
       return 1;
    }

    char i2g_cmd[256];
    readini("I2G_CMD", "att2intl -e", i2g_cmd);
    strcat(i2g_cmd, " " TEMP_DIR "\\");
    strcat(i2g_cmd, SrcFilePrefix);
    strcat(i2g_cmd, ".s");

    if (run(i2g_cmd))
    {
        printf("Error executing asm code convertion command!\n");
        return 1;
    }

    return 0;
}
//---------------------------------------------------------------------------
int getNextUndefSym(FILE * f, char * symbol)
{
    char buf[256];
    char *ptr;
    int i;

    while (!feof(f))
    {
        fgets(buf, sizeof(buf), f);
        if (strstr(buf, "error: symbol"))
        {
            if ((ptr = strchr(buf, ':')) == NULL) continue;
            ptr++;
            if (sscanf(ptr, "%d: error: symbol `%s'", &i, symbol) != 2) continue;
            while (*symbol) if (*symbol == '\'') break; else symbol++;
            *symbol = '\0';
            return i;
        }
    }
    return 0;
}
//---------------------------------------------------------------------------
void putFunName(FILE * f, char * buf)
{
    char nme[128];
    if (*buf == '_') buf++;
    strcpy(nme, buf);
    char * p = strrchr(nme, '@');
    if (p) *p = '\0';
    fprintf(f, "%s\n", nme);
}
//---------------------------------------------------------------------------
// adds function names to a list
int addtoqueue(char * ptr, kolejka * front, kolejka * back)
{
	while (*ptr == ' ' || *ptr == '\t') ptr++; //skip whitespace
	char * t = strrchr(ptr, '*'); //find comment closing
	if (t) *t = '\0'; //erase it
	t = strchr(ptr, '='); //find '='
	if (!t) return 0;
	*t = ' ';
	//devide into two pieces: d1 - function name, d2 - exported function name and push onto queue
	char d1[128], d2[128];
	if (sscanf(ptr, "%s %s", d1, d2) != 2) return 0;
	push(front, back, d1, d2);
    return 1;
}
//---------------------------------------------------------------------------
// post-process assembly
int step2(void)
{
    printf("Step 2 - post-processing assembly\n");
    
    {
        // extract function names [MAKE_EXPORT or MAKE_IMPORT]
        // format: "/* MAKE_EXPORT function_name=export_name */" - C file
        // or: "; MAKE_EXPORT function_name=export_name" - ASM file
        // must be BEFORE the actual definition
        
        char tmp[256];
        int c = 0;
        FILE * input = fopen(SrcFile, "rt");
        while (!feof(input))
        {
            fgets(tmp, sizeof(tmp), input);
            c++;
            char * ptr;
            if ((ptr = strstr(tmp, "MAKE_EXPORT")) != 0)
            {
            	ptr += sizeof("MAKE_EXPORT");
            	if (!addtoqueue(ptr, &exports_front, &exports_back))
                    printf("Warning: Skipping invalid MAKE_EXPORT tag at line %d\n", c);
            }
            else if ((ptr = strstr(tmp, "MAKE_IMPORT")) != 0)
            {
    			ptr += sizeof("MAKE_IMPORT");
            	if (!addtoqueue(ptr, &imports_front, &imports_back))
                    printf("Warning: Skipping invalid MAKE_IMPORT tag at line %d\n", c);
            }
            else if ((ptr = strstr(tmp, "MAKE_CALL")) != 0)
            {
    			ptr += sizeof("MAKE_CALL");
            	if (!addtoqueue(ptr, &calls_front, &calls_back))
                    printf("Warning: Skipping invalid MAKE_CALL tag at line %d\n", c);
            }
        }
        fclose(input);
    }

    char asm_cmd[256];
    readini("NASM_EXEC", "nasm", asm_cmd);
    strcat(asm_cmd, " -E" TEMP_DIR "\\msg.lst " TEMP_DIR "\\");
    strcat(asm_cmd, SrcFilePrefix);
    strcat(asm_cmd, ".asm");

    run(asm_cmd);

    FILE * lst = fopen(TEMP_DIR "\\msg.lst", "rt");
    if (!lst)
    {
        printf("Error opening msg.lst!\n");
        return 1;
    }

    char buf[256], comp[256], symbol[256], *t;
    int n = 1, next;

    strcpy(buf, TEMP_DIR "\\");
    strcat(buf, SrcFilePrefix);
    strcat(buf, ".asm");

    FILE * in = fopen(buf, "rt");

    strcpy(strrchr(buf, '.'), ".new");

    FILE * out = fopen(buf, "wt");
    FILE * out2 = fopen(TEMP_DIR "\\temp.asm", "wt");
    FILE * ffunrels = fopen(TEMP_DIR "\\funrels.txt", "wt");
    FILE * ffunabss = fopen(TEMP_DIR "\\funabss.txt", "wt");

    // append "BITS 32" if needed
    if (fgets(buf, sizeof(buf), in))
    {
        fputs("BITS 32\n", out);
        fputs("BITS 32\norg 0x100\n", out2);
        if (strnicmp(buf, "bits 32", 7))
        {
            fputs(buf, out);
            fputs(buf, out2);
        }
    }

    next = getNextUndefSym(lst, symbol);
    while (fgets(buf, sizeof(buf), in))
    {
        n++;
        if (n == next)
        {
            sscanf(buf, "%s", comp);

            //add "near" to "jmp" if needed
            if (!strcmpi(comp, "jmp") && !strstr(buf, "near"))
            {
                strcpy(comp, buf);
                t = strstr(strlwr(comp), "jmp");
                strcpy(t + 3, " near");
                strcat(t, (char *)((int)buf + t - comp + 3));
                strcpy(buf, comp);
            }

            //check whether it is an absoulute indirect or relative call
            if (*(strstr(buf, symbol) - 1) == '[')
            {
                putFunName(ffunabss, symbol);
                strcpy(strstr(buf, symbol), "$] ;[");
                strcat(buf, symbol);
                strcat(buf, "]\n");
            }
            else
            {
                //try to find an entry on a list
                kolejka t = imports_front;
                if (t) do // if we have overrides
                {
    				char * tmp;
                	if ((tmp = strstr(symbol, t->d1)) != 0)
                	{
                        //check also auto-generated names like "_function@4"
                		if ((tmp == symbol || tmp[-1] == '_') && (*(tmp + strlen(t->d1)) == '\0' || *(tmp + strlen(t->d1)) == '@'))
                		{
    						putFunName(ffunabss, symbol);
    						strcpy(strstr(buf, symbol), "[$] ;[");
    						strcat(buf, symbol);
    						strcat(buf, "]\n");
    						break;
                		}
                	}
                	t = t->next;
                }
                while (t);
            
                if (!t)
                {
                    putFunName(ffunrels, symbol);
                    strcpy(strstr(buf, symbol), "$ ;");
                    strcat(buf, symbol);
                    strcat(buf, "\n");
                }
            }
            next = getNextUndefSym(lst, symbol);
        }
        else
        {
            // skip section/segment tag - this is required by nasm to generate correct list file
            strcpy(comp, buf);
            strlwr(comp);
            if (strstr(comp, "section ") || strstr(comp, "segment ")) continue;
        }

        fputs(buf, out);
        fputs(buf, out2);
    }
    
    fclose(ffunrels);
    fclose(ffunabss);
    fclose(out2);
    fclose(out);
    fclose(in);

    fclose(lst);

    return 0;
}
//---------------------------------------------------------------------------
// create binaries using NASM, calculate addresses
int step3(void)
{
    printf("Step 3 - creating binaries and calculating addresses\n");

    /* prepare binaries for comparison */

    char tmp[256];
    char nasmopti[8];
    readini("NASM_EXEC", "nasm", tmp);
    readini("NASM_OPTI", "99", nasmopti);
    strcat(tmp, " -O");
    strcat(tmp, nasmopti);
    strcat(tmp, " -fbin -l" TEMP_DIR "\\list.lst " TEMP_DIR "\\");
    strcat(tmp, SrcFilePrefix);
    strcat(tmp, ".new");

    if (run(tmp))
    {
        printf("Error compiling asm code! [f1]\n");
        return 1;
    }

    readini("NASM_EXEC", "nasm", tmp);
    readini("NASM_OPTI", "99", nasmopti);
    strcat(tmp, " -O");
    strcat(tmp, nasmopti);
    strcat(tmp, " -fbin " TEMP_DIR "\\temp.asm");

    if (run(tmp))
    {
        printf("Error compiling asm code! [f2]\n");
        return 1;
    }

    /* address comparison */

    // this code locates function (api) addresses which have to be filled with correct values by patcher

    int val1, val2;
    int pos;

    strcpy(tmp, TEMP_DIR "\\");
    strcat(tmp, SrcFilePrefix);

    FILE * funrela = fopen(TEMP_DIR "\\funrela.txt", "wt");
    FILE * funabsa = fopen(TEMP_DIR "\\funabsa.txt", "wt");
    FILE * offs = fopen(TEMP_DIR "\\offs.txt", "wt");

    FILE * in1 = fopen(tmp, "rb");
    FILE * in2 = fopen(TEMP_DIR "\\temp", "rb");

    fread(&val1, sizeof(int), 1, in1);
    fread(&val2, sizeof(int), 1, in2);

    pos = 0;
    
	//read first 4 bytes (32 bit !) to buffer
	{
		//search for FB FF FF FF pattern in 'jmp' and 'call'
		//call/jmp $
		if (val1 == -5 && val2 == -5) fprintf(funrela, "0x%x\n", pos);
		//search for 'push' pattern
		else if (val2 == (val1 + 0x100))
		{
			//push DWORD $
			if ((val1 + 1) == pos) fprintf(funrela, "0x%x\n", pos);
			//call/jmp [$]
			else if ((val1 + 2) == pos) fprintf(funabsa, "0x%x\n", pos);
			//operations involving constants
			else fprintf(offs, "0x%x\n", pos);
		}
	}

	//in loop read next bytes one by one (8 bit) and glue them with previously read 3 bytes
    int v1, v2;
    while ((v1 = fgetc(in1)) != EOF && (v2 = fgetc(in2)) != EOF)
    {
        val1 >>= 8;
        val2 >>= 8;
        ((char *)&val1)[3] = v1;
        ((char *)&val2)[3] = v2;
        pos++;

        if (val1 == -5 && val2 == -5) fprintf(funrela, "0x%x\n", pos);
        else if (val2 == (val1 + 0x100))
        {
            if ((val1 + 1) == pos) fprintf(funrela, "0x%x\n", pos);
            else if ((val1 + 2) == pos) fprintf(funabsa, "0x%x\n", pos);
            else fprintf(offs, "0x%x\n", pos);
        }
    }

    fclose(in2);
    fclose(in1);

    fclose(offs);
    fclose(funrela);
    fclose(funabsa);

    return 0;
}

//---------------------------------------------------------------------------
// create source file(s)
int step4(void)
{
    printf("Step 4 - creating output CPP file\n");

    int c = 0;
    char tmp[256];
    int has_offsets = 0;
    int has_abs_api = 0;
    int has_rel_api = 0;

    // create CPP
    strcpy(tmp, SrcFile);
    char * ptr = strrchr(tmp, '.');
    if (ptr) *ptr = '\0'; else ptr = tmp;
    strcat(ptr, "_out.cpp");

    FILE * out_cpp = fopen(tmp, "wt");

    fprintf(out_cpp, "//---------------------------------------------------------------------------\n"
                     "//%s implementation\n", SrcFilePrefix);
    fprintf(out_cpp, "// Generated by GenPatch v" VERSION " by Xeno86\n");
    {
        char * ptr = strrchr(SrcFile, '\\');
        if (!ptr) ptr = SrcFile; else ptr++;
        fprintf(out_cpp, "// source file: %s\n", ptr);
    }

    //code
    {
        int n = 0;
        strcpy(tmp, TEMP_DIR "\\");
        strcat(tmp, SrcFilePrefix);
        FILE * bin_in = fopen(tmp, "rb");

        fprintf(out_cpp, "static char PI_%s_code[] = {", SrcFilePrefix);

        while ((c = fgetc(bin_in)) != EOF)
        {
            if (!(n % 20)) fprintf(out_cpp, "\n\t");
            fprintf(out_cpp, "0x%02x,", c);
            n++;
        }

        fprintf(out_cpp, "\n};  //code size = 0x%x (%d)\n", n, n);
        fclose(bin_in);
    }

    //fix offsets
    {
        int n = 0;
        FILE * offs = fopen(TEMP_DIR "\\offs.txt", "rt");

        while (fscanf(offs, "0x%x\n", &c) != EOF)
        {
            //open bracket...
            if (!has_offsets)
            {
                fprintf(out_cpp, "static DWORD PI_%s_ofs[] = {", SrcFilePrefix);
                has_offsets = 1;
            }

            if (!(n % 10)) fprintf(out_cpp, "\n\t");
            fprintf(out_cpp, "0x%02x, ", c);
            n++;
        }

        //close bracket...
        if (has_offsets)
        {
            fprintf(out_cpp, "0,\n};\n");
        }

        fclose(offs);
    }

    //fix absoulute addresses
    {
        FILE * funabsa = fopen(TEMP_DIR "\\funabsa.txt", "rt");
        FILE * funabss = fopen(TEMP_DIR "\\funabss.txt", "rt");

        while (fscanf(funabss, "%s", tmp) != EOF && fscanf(funabsa, "0x%x\n", &c) != EOF)
        {
            //skip "DUMMY" keyword
			if (!strcmp(tmp, DUMMY_KEYWORD)) continue;
			
            //open bracket...
            if (!has_abs_api)
            {
                fprintf(out_cpp, "static api_entry PI_%s_abs[] = {\n", SrcFilePrefix);
                has_abs_api = 1;
            }

            // TODO: add Export/Import detection
            //try to find an entry on a list
            kolejka t = imports_front;
            if (t) do // if we have overrides
            {
            	if (!strcmp(tmp, t->d1))
            	{
            		strcpy(tmp, t->d2);
            		break;
            	}
            	t = t->next;
            }
            while (t);
            fprintf(out_cpp, "\t{ 0x%02x, \"%s\" },\n", c, tmp);
        }

        //close bracket...
        if (has_abs_api)
        {
            fprintf(out_cpp, "\t{ 0, \"\" },\n"
                             "};\n");
        }

        fclose(funabsa);
        fclose(funabss);
    }

    //fix relative addresses
    {
        FILE * funrela = fopen(TEMP_DIR "\\funrela.txt", "rt");
        FILE * funrels = fopen(TEMP_DIR "\\funrels.txt", "rt");

        while (fscanf(funrels, "%s", tmp) != EOF && fscanf(funrela, "0x%x\n", &c) != EOF)
        {
            //skip "DUMMY" keyword
			if (!strcmp(tmp, DUMMY_KEYWORD)) continue;
			
            //open bracket...
            if (!has_rel_api)
            {
                fprintf(out_cpp, "static api_entry PI_%s_rel[] = {\n", SrcFilePrefix);
                has_rel_api = 1;
            }

            // TODO: add Export/Import detection
            //try to find an entry on a list
            kolejka t = calls_front;
            if (t) do // if we have overrides
            {
            	if (!strcmp(tmp, t->d1))
            	{
            		strcpy(tmp, t->d2);
            		break;
            	}
            	t = t->next;
            }
            while (t);
            fprintf(out_cpp, "\t{ 0x%02x, \"%s\" },\n", c, tmp);
        }

        //close bracket...
        if (has_rel_api)
        {
            fprintf(out_cpp, "\t{ 0, \"\" },\n"
                             "};\n");
        }

        fclose(funrela);
        fclose(funrels);
    }

    //hook api
    // TODO: special flags to C for identification (?)
    {
		int has_hook_api = 0;
        fprintf(out_cpp, "static api_entry PI_%s_hook[] = {\n", SrcFilePrefix);

        FILE * list = fopen(TEMP_DIR "\\list.lst", "rt");
        while (exports_front && !feof(list))
        {
            fgets(tmp, sizeof(tmp), list);
            char * ptr;
            if (strlen(tmp) > 41) ptr = tmp + 40; else continue;
            //find & remove ':'
            char * t = strchr(ptr, ':');
            if (!t) continue;
            *t = '\0';
            //remove '@'
            t = strchr(ptr, '@');
            if (t) *t = '\0';
            if (*ptr == '_') ptr++;
            //compare with queue
            if (strcmp(ptr, exports_front->d1)) continue;
            int found = 0;
            //skip until we reach line with address
            while (!feof(list))
            {
                fgets(tmp, sizeof(tmp), list);
                if (tmp[7] == ' ') continue;
                found = 1;
                break;
            }
            if (!found)
            {
                printf("Error: Couldn't find matching definiton for %s!\n", exports_front->d1);
                if (!stricmp(Ext, "asm")) { printf("Hint: Missing @ sign?\n"); }
                fprintf(out_cpp, "ERROR!!!\n");
                fclose(list);
                fclose(out_cpp);
                return 1;
            }

            int i;
            sscanf(tmp + 7, "%08x", &i);

            fprintf(out_cpp, "\t{ 0x%02x, \"%s\" },\n", i, exports_front->d2);
            fseek(list, 0, SEEK_SET);

            pop(&exports_front, &exports_back);
            has_hook_api = 1;
        }
        if (!has_hook_api) printf("Warning: Hook api list is empty! Missing MAKE_EXPORT tag?\n");
        fprintf(out_cpp, "\t{ 0, \"\" },\n"
                         "};\n");

        fclose(list);
    }

    //patch_item declaration
    fprintf(out_cpp, "static patch_item PI_%s(PI_%s_code, sizeof(PI_%s_code), ", SrcFilePrefix, SrcFilePrefix, SrcFilePrefix);
    if (has_offsets) fprintf(out_cpp, "PI_%s_ofs, ", SrcFilePrefix); else fprintf(out_cpp, "NULL, ");
    if (has_abs_api) fprintf(out_cpp, "PI_%s_abs, ", SrcFilePrefix); else fprintf(out_cpp, "NULL, ");
    if (has_rel_api) fprintf(out_cpp, "PI_%s_rel, ", SrcFilePrefix); else fprintf(out_cpp, "NULL, ");
    fprintf(out_cpp, "PI_%s_hook, NULL);\n\n", SrcFilePrefix);

    fclose(out_cpp);

    return 0;
}
//---------------------------------------------------------------------------
void __cdecl exit_func(void)
{
    //free the queue
    while (imports_front) pop(&imports_front, &imports_back);
    while (exports_front) pop(&exports_front, &exports_back);
    while (calls_front) pop(&calls_front, &calls_back);

    //remove temporary files
    remove(TEMP_DIR "\\funrela.txt"); //relative address call/jmp $
    remove(TEMP_DIR "\\funrels.txt"); //relative address function name
    remove(TEMP_DIR "\\funabsa.txt"); //absolute address (pointer) call/jmp [$]
    remove(TEMP_DIR "\\funabss.txt"); //absolute address function name
    remove(TEMP_DIR "\\list.lst");    //assembly file w/offsets
    remove(TEMP_DIR "\\msg.lst");     //NASM output messages
    remove(TEMP_DIR "\\offs.txt");    //absolute address (variable) push $
    remove(TEMP_DIR "\\temp.asm");    //postprocessed assembly w/nops
    remove(TEMP_DIR "\\temp");        // ... compiled

    char tmp[MAX_PATH];
    strcpy(tmp, TEMP_DIR "\\");
    strcat(tmp, SrcFilePrefix);
    remove(tmp);                      //compiled assembly
    strcat(tmp, ".new");
    remove(tmp);                      //postprocessed assembly
    strcpy(tmp, TEMP_DIR "\\");
    strcat(tmp, SrcFilePrefix);
    strcat(tmp, ".asm");
    remove(tmp);                      //assembled source
    strcpy(tmp, TEMP_DIR "\\");
    strcat(tmp, SrcFilePrefix);
    strcat(tmp, ".s");
    remove(tmp);                      //assembled source (ATT format)
    
    RemoveDirectory(TEMP_DIR);        //remove directory if empty
}

//---------------------------------------------------------------------------

void process_commandline(int argc, char *argv[])
{
    int i = 1;

    if (argc > 1) while (argv[i][0] == '-' && i < argc)
    {
        if      (!strcmp(argv[i]+1, "dbg")) param_debug = 1;
        else
        {
            printf("Invalid parameter: '%s'\n", argv[i]);
            exit(1);
        }
        i++;
    }
    if (i >= argc)
    {
       printf("No file specified!\n");
       exit(1);
    }
    GetFullPathName(argv[i], MAX_PATH, SrcFile, NULL);
}

//---------------------------------------------------------------------------

int main(int argc, char **argv)
{
    printf("GenPatch v" VERSION " by Xeno86\n\n");

    process_commandline(argc, argv);

    //change current directory to find tools
    char ExeDir[MAX_PATH];
    GetModuleFileName(NULL, ExeDir, MAX_PATH);
    char * ptr = strrchr(ExeDir, '\\');
    *ptr = '\0';
    SetCurrentDirectory(ExeDir);

    if (!param_debug) atexit(exit_func);

    ptr = strrchr(SrcFile, '\\');
    if (!ptr) ptr = SrcFile; else ptr++;

    strcpy(SrcFilePrefix, ptr);
    char * e = strrchr(SrcFilePrefix, '.');
    if (!e)
    {
		printf("Invalid input file!\n");
		exit(1);
    }
    *(e++) = '\0';
    strncpy(Ext, e, 3);

    strcpy(confFile, argv[0]);
    confFile[strlen(confFile) - 4] = '\0';
    strcat(confFile, ".ini");

    if (step0() || step1() || step2() || step3() || step4())
    {
        printf("Error has occured!\n");
        return 1;
    }

    printf("All ok!\n");

    return 0;
}
