/*
 *  ModSecAttr
 *  Copyright (C) 2007, Xeno86
 *
 *  This file is part of KernelEx source code.
 *
 *  KernelEx is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; version 2 of the License.
 *
 *  KernelEx is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include "pefile.h"
#include <string.h>

unsigned long decodeCharacteristics(const char arg[])
{
	unsigned long ret = 0;
	
	if (!strcmpi(arg, "READ")) ret = IMAGE_SCN_MEM_READ;
	else if (!strcmpi(arg, "WRITE")) ret = IMAGE_SCN_MEM_WRITE;
	else if (!strcmpi(arg, "EXECUTE")) ret = IMAGE_SCN_MEM_EXECUTE;
	else if (!strcmpi(arg, "SHARED")) ret = IMAGE_SCN_MEM_SHARED;
	else
	{
		printf("ERROR: Not supported characteristics: %s\n", arg);
		exit(1);
	}
	
	return ret;
}

int main(int argc, char ** argv)
{
	printf("ModSecAttr - Copyright (C) 2007, Xeno86\n\n");
	if (argc <= 2)
	{
		printf("ERROR: Too few arguments\n");
		printf("\nUsage:\t%s file section [+attr1] [-attr2] ... [+-attrN]\n\n", strrchr(argv[0], '\\') + 1);
		printf("  file         Target executable\n");
		printf("  section      Section in that file\n");
		printf("  attrN        One of the attributes: READ, WRITE, EXECUTE, SHARED\n");
		printf("               + enables; - disables the attribute\n");
		return 1;
	}
	
	CPEFile file;
	if (!file.LoadFromFile(argv[1]))
	{
		printf("ERROR: Error loading file %s\n", argv[1]);
		return 1;
	}
	
	IMAGE_SECTION_HEADER* sechdr = file.GetSection(argv[2]);
	
	if (!sechdr)
	{
		printf("ERROR: Section %s not found\n", argv[2]);
		return 1;
	}
	
	unsigned long characteristics = sechdr->Characteristics;
	
	printf("Current file characteristics: 0x%08lx\n", characteristics);
	
	for (int i = 3 ; i < argc ; i++)
	{
		if (argv[i][0] == '+')
		{
			characteristics |= decodeCharacteristics(argv[i] + 1);
		}
		else if (argv[i][0] == '-')
		{
			characteristics &= ~decodeCharacteristics(argv[i] + 1);
		}
		else
		{
			printf("ERROR: Invalid format\n");
			return 1;
		}
	}
	
	if (sechdr->Characteristics != characteristics)
	{	
		printf("New file characteristics:     0x%08lx\n", characteristics);
		sechdr->Characteristics = characteristics;
		
		file.FixChecksum();
		
		if (!file.SaveToFile(argv[1]))
		{
			printf("ERROR: Error saving file %s\n", argv[1]);
			return 1;
		}
	}
	else
	{
		printf("No changes\n");
	}
	
	return 0;
}
